/* -*- indent-tabs-mode: t; tab-width: 4; c-basic-offset: 4; -*- */
/*
   hashing_struct.c
   Projet GtkCmpHash

   (C) Copyright 2007 Olivier Delhomme
   e-mail : olivierdelhomme@gmail.com
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.           
*/
/**
 *  Gestion du chargement, de la mise en mmoire et des comparaisons
 *  des hashs
 *  Utilisation de la structure :
 *   - Cration de l'arbre/tronc : nouveau_tronc(niveau)
 *     niveau indique le nombre d'indirections (intressant  partir de 2)
 *   - Insertion dans l'arbre/tronc : insere_dans_tronc(tronc, file_hash)
 *     Insre file_hash dans le tronc (l o il faut)
 *   - Recherche : recherche_dans_tronc(tronc, file_hash)
 *     Recherche file_hash rapidement (en quelques comparaisons) avec 
 *     gestion des collisions
 */

#include "gtkcmphash.h"

static arbre_t *nouvel_arbre(guint niveau_max, guint niveau);
static guint cle_de_hashage(guchar c);
static guint int_value_of_hex(guchar c);
static GSList *insere_dans_liste(GSList *list, file_hash_t *file_hash);
static void insere_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash);
static GSList * recherche_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash);
static GSList *recherche_dans_liste(GSList *list, file_hash_t *file_hash);
static gint my_g_ascii_strcasecmp (const guchar *s1, const guchar *s2, guint len);

/**
 *  Cre un nouvel arbre de profondeur "niveau_max",  partir de "niveau"
 */
static arbre_t *nouvel_arbre(guint niveau_max, guint niveau)
{
  arbre_t *arbre = NULL;
  guint i = 0;

  arbre = (arbre_t *) g_malloc0(sizeof(arbre_t));
  arbre->niveau = niveau;
  
  if (niveau >= niveau_max)
    { 
      for (i=0; i<16; i++)
	arbre->array[i] = NULL;
      return arbre;
    }
  else
    {
      for (i=0; i<16; i++)
	arbre->array[i] = nouvel_arbre(niveau_max, niveau+1);
      return arbre;
    }
}

/**
 *  Construit une structure vide de niveau "niveau"
 *  maximum 16 sous niveaux pour le md5 
 */
hash_t *nouveau_tronc(guint niveau_max)
{
  hash_t *tronc = NULL;

  tronc = (hash_t *) g_malloc0(sizeof(hash_t));

  tronc->arbre = nouvel_arbre(niveau_max, 1);
  tronc->nb_niveau = niveau_max;

  return tronc;
}

/**
 *  Cl de hashage ( partir d'un octet renvoi 16
 *  cas possibles)
 */
static guint cle_de_hashage(guchar c)
{
  return ((guint)(c >>4));
}

/**
 *  Retourne la valeur du caractre hxadcimal en dcimal
 *  0->0 ... 9->9, a->10, ... f->15 sinon 0
 */
static guint int_value_of_hex(guchar c)
{
  switch (c)
    {
    case '0':
      return 0;
      break;
    case '1':
      return 1;
      break;
    case '2':
      return 2;
      break;
    case '3':
      return 3;
      break;
    case '4':
      return 4;
      break;
    case '5':
      return 5;
      break;
    case '6':
      return 6;
      break;
    case '7':
      return 7;
      break;
    case '8':
      return 8;
      break;
    case '9':
      return 9;
      break;
    case 'a':
      return 10;
      break;
    case 'b':
      return 11;
      break;
    case 'c':
      return 12;
      break;
    case 'd':
      return 13;
      break;
    case 'e':
      return 14;
      break;
    case 'f':
      return 15;
      break;
    default:
      return 0;
      break;
    }
}

/**
 *  Insre le file_hash d'un fichier dans une liste
 *  cre ventuellement cette liste
 */
static GSList *insere_dans_liste(GSList *list, file_hash_t *file_hash)
{
  list = g_slist_prepend(list, file_hash);

  return list;
}

/**
 *  Insre un file_hash d'un fichier dans l'arbre (en fonction de son md5)
 *  TODO : laisser le choix  l'utilisateur du hash servant de comparaison
 */
static void insere_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash)
{
  guint numero_liste = 0;
  guchar *hash = file_hash->hash_md5;

  numero_liste = cle_de_hashage(hash[arbre->niveau-1]);

  if (arbre->niveau >= niveau_max)
    arbre->array[numero_liste] = insere_dans_liste(arbre->array[numero_liste], file_hash); 
  else  
    insere_dans_arbre(arbre->array[numero_liste], niveau_max, file_hash); 
}

/**
 *  Insre un file_hash d'un fichier dans la structure
 */
void insere_dans_tronc(hash_t *tronc, file_hash_t *file_hash)
{
  
  if (tronc != NULL)
    {
      if (tronc->arbre != NULL && tronc->nb_niveau > 0)
	{
	  insere_dans_arbre(tronc->arbre, tronc->nb_niveau, file_hash);
	}
    }
}

/**
 *  Insre toute une liste de file_hash dans la structure tronc
 */
void insere_la_liste_dans_tronc(hash_t *tronc, GSList *file_hash_list)
{
  file_hash_t *file_hash = NULL;

  while (file_hash_list != NULL)
    {
      file_hash = (file_hash_t *) file_hash_list->data;

      if (file_hash != NULL)
	insere_dans_tronc(tronc, file_hash);

      file_hash_list = g_slist_next(file_hash_list);
    }
}

/**
 *  Procdure copie de la glib, mais dans laquelle
 *  j'ai enlev les TOLOWER dont je n'ai absolument 
 *  pas besoin... donc gain de rapidit (c'est plus rapide que memcmp)
 */
static gint my_g_ascii_strcasecmp(const guchar *s1, const guchar *s2, guint len)
{

  while (len)
    {
      if ((gint)(guchar)(*s1) != (gint)(guchar)(*s2))
	return (((gint)(guchar) *s1) - ((gint)(guchar) *s2));
	
      s1++; 
      s2++; 
      len--;
    }

  return (((gint)(guchar) *s1) - ((gint)(guchar) *s2));
}

/**
 *  Recherche dans la liste d'ventuels fichiers ayant un hash 
 *  identique  celui de file_hash (utilise le md5)
 */
static GSList *recherche_dans_liste(GSList *list, file_hash_t *file_hash)
{
   file_hash_t *tmp_hash = NULL;
   GSList *head = list;
   GSList *file_hash_list = NULL;

  while (list != NULL)
    {
      tmp_hash = list->data;
      
      if (tmp_hash != NULL)
	{
	  if (my_g_ascii_strcasecmp(file_hash->hash_md5, tmp_hash->hash_md5, tmp_hash->len_md5) == 0) 
	    file_hash_list = g_slist_prepend(file_hash_list, tmp_hash);	
	}

      list = list->next;
    }
  
  list = head;
  
  return file_hash_list;
}

/**
 *  Effectue la recherche dans un arbre
 */
static GSList *recherche_dans_arbre(arbre_t *arbre, guint niveau_max, file_hash_t *file_hash)
{
  guint numero_liste = 0;
  guchar *hash = file_hash->hash_md5;

  /* il faut utiliser la mme clef que celle de l'insertion */
  /* TODO : donner le choix de la clef  l'utilisateur      */
  numero_liste = cle_de_hashage(hash[arbre->niveau-1]);

  if (arbre->niveau >= niveau_max)
    return recherche_dans_liste(arbre->array[numero_liste], file_hash);
  else
    return recherche_dans_arbre(arbre->array[numero_liste], niveau_max, file_hash);
}

/**
 *  Recherche si le hash "hash" est prsent dans la structure de donnes
 *  la fonction retourne la liste des hash correspondant
 *  NULL dans les autres cas
 */ 
GSList *recherche_dans_tronc(hash_t *tronc, file_hash_t *file_hash)
{
  if (tronc != NULL && tronc->arbre != NULL && tronc->nb_niveau > 0)
    return recherche_dans_arbre(tronc->arbre, tronc->nb_niveau, file_hash);
  else
    return NULL;
}

/**
 *  Procdure qui transforme le hash b1f548... en son
 *  quivalent binaire; Ralise un g_malloc0 qu'on pourra
 *  librer avec un g_free par la suite.
 */
guchar *transforme_le_hash_de_hex_en_binaire(guchar *hash)
{
  guchar c = (guchar) 0;
  guint i = 0;
  guint j = 0;
  guint len = strlen((const char *)hash);
  guchar *aux = NULL;

  aux = (guchar *) g_malloc0(sizeof(guchar)*((len/2)+2));
  
  while (i < len)
    {
      c = 16*int_value_of_hex(hash[i]) + int_value_of_hex(hash[i+1]);
      aux[j] = c;
      j++;
      i +=2;
    }
  aux[j] = (guchar) 0;

  return aux;
}

/**
 *  Fonction qui transforme un hash sous forme binaire
 *  en un hash sous forme hexadcimale affichable "12f5cda5..."
 *  Ralise un g_malloc0 qu'on pourra librer par la suite avec
 *  un g_free
 */
guchar *transforme_le_hash_de_binaire_en_hex(guchar *hash, guint len)
{
  guint i = 0;
  guchar *aux = NULL;
  guchar *md5 = NULL;

  aux = (guchar *) g_malloc0(sizeof(gchar)*3);
  md5 = (guchar *) g_malloc0(sizeof(guchar)*((len*2)+2));

   for(i = 0; i < len; i++)  
    {
		sprintf((char *) aux, "%02x", hash[i]); 
      memcpy(md5+(i*2), aux, 2);
    }
   md5[i*2+1] = '\0';
   g_free(aux);

   return md5;
}

/**
 *  Fonction de libration de la liste
 */
void free_file_hash_list(GSList *file_hash_list)
{
  GSList *head = file_hash_list;
  file_hash_t *file_hash = NULL;

  while (file_hash_list != NULL)
    {
      file_hash = (file_hash_t *) file_hash_list->data;
      
      if (file_hash != NULL)
	{
	  if (file_hash->hash_md5 != NULL)
	    g_free(file_hash->hash_md5);
	  if (file_hash->hash_sha1 != NULL)
	    g_free(file_hash->hash_sha1);
	  if (file_hash->hash_ripemd != NULL)
	    g_free(file_hash->hash_ripemd);
	  if (file_hash->filename != NULL)
	    g_free(file_hash->filename);
	  if (file_hash->hashset_name != NULL)
	    g_free(file_hash->hashset_name);
	}
      file_hash_list = g_slist_next(file_hash_list);
    }
  
  g_slist_free(head);
}

/**
 *  Fonction pour trouver les diffrences entre une liste de fichiers
 *  hachs et un hashset chargs en mmoire
 */
found_or_not_t *find_all_hashes_from_hashset(GSList *file_hash_list, hash_t *tronc)
{
  file_hash_t *file_hash = NULL; /* pour parcourir la liste des fichiers hachs */
  file_hash_t *result_hash = NULL;
  GSList *result = NULL;  /* rsultat de la recherche dans l'arbre */
  found_or_not_t *dedans_ou_pas = NULL;

  dedans_ou_pas = (found_or_not_t *) g_malloc0(sizeof(found_or_not_t));
  dedans_ou_pas->found = NULL;
  dedans_ou_pas->not_found = NULL;
  
  while (file_hash_list != NULL)
    {
      file_hash = (file_hash_t *) file_hash_list->data;

      if (file_hash != NULL)
	{
	  result = recherche_dans_tronc(tronc, file_hash);

	  if (result != NULL)
	    {
	      result_hash = (file_hash_t *) result->data; /* on prend le premier on pourrait tous les prendre */
	      file_hash->hashset_name = result_hash->hashset_name;
	      dedans_ou_pas->found = g_slist_prepend(dedans_ou_pas->found, file_hash);
	    }
	  else
	    dedans_ou_pas->not_found = g_slist_prepend(dedans_ou_pas->not_found, file_hash);

	  g_slist_free(result);
	}

      file_hash_list = g_slist_next(file_hash_list);
    }
  
  return dedans_ou_pas;
}
